@@ -15,7 +15,7 @@ import Data.List.NonEmpty (NonEmpty)
 import Data.Map.Strict qualified as Map
 import Data.Maybe (fromMaybe, mapMaybe, maybeToList)
 import Data.Set qualified as Set
-import Data.Text.IO qualified as TIO
+import Data.Text.IO.Utf8 qualified as T.Utf8
 import Data.Version (showVersion)
 import Data.Yaml qualified as Yaml
 import Distribution.ModuleName (ModuleName)
@@ -30,8 +30,8 @@ import Ormolu.Parser (manualExts)
 import Ormolu.Terminal
 import Ormolu.Utils (showOutputable)
 import Ormolu.Utils.Fixity
-import Ormolu.Utils.IO
 import Paths_fourmolu (version)
+import System.Console.Terminal.Size qualified as Terminal
 import System.Directory
 import System.Exit (ExitCode (..), exitWith)
 import System.FilePath qualified as FP
@@ -40,13 +40,8 @@ import System.IO (hPutStrLn, stderr)
 -- | Entry point of the program.
 main :: IO ()
 main = do
-  opts@Opts {..} <- execParser optsParserInfo
-
-  cwd <- getCurrentDirectory
-  cfg <- case optInputFiles of
-    [] -> mkConfig cwd opts
-    ["-"] -> mkConfig cwd opts
-    file : _ -> mkConfig file opts
+  opts@Opts {..} <- runParser optsParserInfo
+  cfg <- resolveConfig opts
 
   let formatOne' =
         formatOne
@@ -75,27 +70,34 @@ main = do
                 else 102
 
   exitWith exitCode
+  where
+    runParser p = do
+      termWidth <- maybe 100 Terminal.width <$> Terminal.size
+      customExecParser (prefs $ helpIndent 35 <> columns termWidth) p
 
 -- | Build the full config, by adding 'PrinterOpts' from a file, if found.
-mkConfig :: FilePath -> Opts -> IO (Config RegionIndices)
-mkConfig path Opts {optQuiet, optConfig = cliConfig, optPrinterOpts = cliPrinterOpts} = do
+resolveConfig :: Opts -> IO (Config RegionIndices)
+resolveConfig opts@(Opts {optConfig = cliConfig, optPrinterOpts = cliPrinterOpts}) = do
   fourmoluConfig <-
-    loadConfigFile path >>= \case
-      ConfigLoaded f cfg -> do
-        outputInfo $ "Loaded config from: " <> f
-        outputDebug $ unwords ["*** CONFIG FILE ***", show cfg]
-        pure cfg
-      ConfigParseError f e -> do
-        outputError . unlines $
-          [ "Failed to load " <> f <> ":",
-            Yaml.prettyPrintParseException e
-          ]
-        exitWith $ ExitFailure 400
-      ConfigNotFound searchDirs -> do
+    findConfigFile' >>= \case
+      Left ConfigNotFound {searchDirs} -> do
         outputDebug . unlines $
           ("No " ++ show configFileName ++ " found in any of:")
             : map ("  " ++) searchDirs
         pure emptyConfig
+      Right configPath ->
+        Yaml.decodeFileEither configPath >>= \case
+          Left e -> do
+            outputError . unlines $
+              [ "Failed to load " <> configPath <> ":",
+                Yaml.prettyPrintParseException e
+              ]
+            exitWith $ ExitFailure 400
+          Right cfg -> do
+            outputInfo $ "Loaded config from: " <> configPath
+            outputDebug $ unwords ["*** CONFIG FILE ***", show cfg]
+            pure cfg
+
   return $
     cliConfig
       { cfgPrinterOpts =
@@ -115,9 +117,17 @@ mkConfig path Opts {optQuiet, optConfig = cliConfig, optPrinterOpts = cliPrinter
             ]
       }
   where
+    findConfigFile' = do
+      cwd <- getCurrentDirectory
+      case optInputFiles opts of
+        _ | Just configPath <- optConfigFilePath opts -> pure $ Right configPath
+        [] -> findConfigFile cwd
+        ["-"] -> findConfigFile cwd
+        file : _ -> findConfigFile $ FP.takeDirectory file
+
     output = hPutStrLn stderr
     outputError = output
-    outputInfo = unless optQuiet . output
+    outputInfo = unless (optQuiet opts) . output
     outputDebug = when (cfgDebug cliConfig) . output
 
 getHaskellFiles :: FilePath -> IO [FilePath]
@@ -191,7 +201,7 @@ formatOne ConfigFileOpts {..} mode reqSourceType rawConfig mpath =
         config <- patchConfig Nothing mcabalInfo mdotOrmolu
         case mode of
           Stdout -> do
-            ormoluStdin config >>= TIO.putStr
+            ormoluStdin config >>= T.Utf8.putStr
             return ExitSuccess
           InPlace -> do
             hPutStrLn
@@ -201,7 +211,7 @@ formatOne ConfigFileOpts {..} mode reqSourceType rawConfig mpath =
             return (ExitFailure 101)
           Check -> do
             -- ormoluStdin is not used because we need the originalInput
-            originalInput <- getContentsUtf8
+            originalInput <- T.Utf8.getContents
             let stdinRepr = "<stdin>"
             formattedInput <-
               ormolu config stdinRepr originalInput
@@ -220,19 +230,19 @@ formatOne ConfigFileOpts {..} mode reqSourceType rawConfig mpath =
             mdotOrmolu
         case mode of
           Stdout -> do
-            ormoluFile config inputFile >>= TIO.putStr
+            ormoluFile config inputFile >>= T.Utf8.putStr
             return ExitSuccess
           InPlace -> do
             -- ormoluFile is not used because we need originalInput
-            originalInput <- readFileUtf8 inputFile
+            originalInput <- T.Utf8.readFile inputFile
             formattedInput <-
               ormolu config inputFile originalInput
             when (formattedInput /= originalInput) $
-              writeFileUtf8 inputFile formattedInput
+              T.Utf8.writeFile inputFile formattedInput
             return ExitSuccess
           Check -> do
             -- ormoluFile is not used because we need originalInput
-            originalInput <- readFileUtf8 inputFile
+            originalInput <- T.Utf8.readFile inputFile
             formattedInput <-
               ormolu config inputFile originalInput
             handleDiff originalInput formattedInput inputFile
@@ -274,6 +284,8 @@ data Opts = Opts
     optQuiet :: !Bool,
     -- | Ormolu 'Config'
     optConfig :: !(Config RegionIndices),
+    -- | Ormolu 'Config'
+    optConfigFilePath :: !(Maybe FilePath),
     -- | Fourmolu 'PrinterOpts',
     optPrinterOpts :: PrinterOptsPartial,
     -- | Options related to info extracted from files
@@ -363,6 +375,11 @@ optsParser =
         help "Make output quieter"
       ]
     <*> configParser
+    <*> (optional . strOption . mconcat)
+      [ metavar "CONFIG_FILE",
+        long "config",
+        help "Path to the config file to use. If not specified, tries to discover one automatically."
+      ]
     <*> printerOptsParser
     <*> configFileOptsParser
     <*> sourceTypeParser
@@ -456,7 +473,13 @@ configParser =
                 help "End line of the region to format (inclusive)"
               ]
         )
-    <*> pure defaultPrinterOpts -- unused; overwritten in mkConfig
+    <*> pure defaultPrinterOpts -- unused; overwritten in resolveConfig
+    <*> (fmap Set.fromList . many . strOption . mconcat)
+      [ long "local-modules",
+        short 'm',
+        metavar "LOCAL_MODULES",
+        help "Modules Fourmolu should consider as local by the current Cabal package"
+      ]
 
 sourceTypeParser :: Parser (Maybe SourceType)
 sourceTypeParser =
