@@ -9,9 +9,7 @@ data Foo a where
         forall a b.
         (Show a, Eq b) => -- foo
         -- bar
-        a ->
-        b ->
-        Foo 'Int
+        a -> b -> Foo 'Int
     -- | But 'Bar' is also not too bad.
     Bar ::
         Int -> Maybe Text -> Foo 'Bool
