@@ -7,7 +7,6 @@ instance (Eq a) => Eq (Foo a) where
 
 instance (Num r, V.Vector v r, Factored m r) => Num (VT v m r) where
     {-# SPECIALIZE instance
-        ( (Factored m Int) => Num (VT U.Vector m Int)
-        )
+        ((Factored m Int) => Num (VT U.Vector m Int))
         #-}
     VT x + VT y = VT $ V.zipWith (+) x y
