@@ -6,5 +6,4 @@ wrapError ::
       HasCatch innertag inner (t m')
     , HasCatch outertag outer m
     ) =>
-    (forall m'. (HasCatch innertag inner m') => m' a) ->
-    m a
+    (forall m'. (HasCatch innertag inner m') => m' a) -> m a
