@@ -24,6 +24,20 @@ barbaz x y z w =
     ]
 
 a = do
-    [ c
-        | c <- d
+    d <-
+        [ x + 1
+        | x <- b
         ]
+
+    [ c
+      | c <- d
+      ]
+
+trans =
+    [ x
+    | x <- xs
+    , then
+        reverse
+    , then
+        reverse
+    ]
