@@ -3,8 +3,9 @@ module Foo
 
       -- * Something
       bar,
-      -- | A multiline
-      -- comment here
+      {- | A multiline
+      comment here
+      -}
       baz,
 
       -- * Another thing
@@ -50,3 +51,19 @@ module Foo
       -- * Some other thing
       Foo,
     ) where
+
+{- // -}
+
+-- See data/examples/import/docstrings-after-exports.hs
+module Test
+    ( since1, -- ^ @since 1.0
+      since2, -- ^ @since 2.0
+      since3, -- ^ @since 3.0
+      SinceType (..), -- ^ @since 4.0
+      SinceClass (..), -- ^ @since 5.0
+      Multi (..),
+      {- ^ since 6.0
+      multi
+      line
+      -}
+    ) where
