@@ -3,8 +3,10 @@
 module Foo (
   foo, bar, baz) where
 
+import Control.Category
+
 foo :: Int
-foo = 5
+foo  =  5
 
 bar :: Int -> Int
 bar = \case
@@ -14,4 +16,8 @@ bar = \case
 baz :: Int -> Int
 baz = gege
   where
-    gege = 1   + 2
+    gege =
+      1
+        >>> zeze
+          + 2
+        >>> nona
