@@ -7,6 +7,8 @@ module Foo
   )
 where
 
+import Control.Category
+
 foo :: Int
 foo = 5
 
@@ -18,4 +20,8 @@ bar = \case
 baz :: Int -> Int
 baz = gege
   where
-    gege = 1 + 2
+    gege =
+      1
+        >>> zeze
+          + 2
+        >>> nona
