@@ -1,3 +1,4 @@
+{-# LANGUAGE DataKinds #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE OverloadedStrings #-}
 
@@ -12,11 +13,13 @@ module Ormolu.Printer.Internal
 
     -- * Internal functions
     txt,
+    txtStripIndent,
     interferingTxt,
     atom,
     space,
     newline,
     declNewline,
+    multilineCommentNewline,
     askSourceType,
     askModuleFixityMap,
     askDebug,
@@ -30,6 +33,7 @@ module Ormolu.Printer.Internal
     vlayout,
     getLayout,
     getPrinterOpt,
+    getLocalModules,
 
     -- * Helpers for braces
     useBraces,
@@ -42,6 +46,7 @@ module Ormolu.Printer.Internal
     trimSpanStream,
     nextEltSpan,
     popComment,
+    getEnclosingComments,
     getEnclosingSpan,
     getEnclosingSpanWhere,
     withEnclosingSpan,
@@ -63,14 +68,20 @@ import Control.Monad
 import Control.Monad.Reader
 import Control.Monad.State.Strict
 import Data.Bool (bool)
+import Data.Char (isSpace)
+import Data.Choice (Choice)
+import Data.Choice qualified as Choice
 import Data.Coerce
+import Data.Functor ((<&>))
 import Data.Functor.Identity (runIdentity)
 import Data.List (find)
 import Data.Maybe (listToMaybe)
+import Data.Set (Set)
 import Data.Text (Text)
 import Data.Text qualified as T
 import Data.Text.Lazy qualified as TL
 import Data.Text.Lazy.Builder
+import Distribution.ModuleName (ModuleName)
 import GHC.Data.EnumSet (EnumSet)
 import GHC.Data.EnumSet qualified as EnumSet
 import GHC.LanguageExtensions.Type
@@ -103,6 +114,7 @@ data RC = RC
     -- | Whether the last expression in the layout can use braces
     rcCanUseBraces :: Bool,
     rcPrinterOpts :: PrinterOptsTotal,
+    rcLocalModules :: Set ModuleName,
     -- | Enabled extensions
     rcExtensions :: EnumSet Extension,
     -- | Whether the source is a signature or a regular module
@@ -175,6 +187,7 @@ runR ::
   -- | Comment stream
   CommentStream ->
   PrinterOptsTotal ->
+  Set ModuleName ->
   -- | Whether the source is a signature or a regular module
   SourceType ->
   -- | Enabled extensions
@@ -184,7 +197,7 @@ runR ::
   -- | Resulting rendition
   Bool ->
   Text
-runR (R m) sstream cstream printerOpts sourceType extensions moduleFixityMap debug =
+runR (R m) sstream cstream printerOpts localModules sourceType extensions moduleFixityMap debug =
   TL.toStrict . toLazyText . scBuilder $ execState (runReaderT m rc) sc
   where
     rc =
@@ -194,6 +207,7 @@ runR (R m) sstream cstream printerOpts sourceType extensions moduleFixityMap deb
           rcEnclosingSpans = [],
           rcCanUseBraces = False,
           rcPrinterOpts = printerOpts,
+          rcLocalModules = localModules,
           rcExtensions = extensions,
           rcSourceType = sourceType,
           rcModuleFixityMap = moduleFixityMap,
@@ -245,6 +259,15 @@ txt ::
   R ()
 txt = spit SimpleText
 
+-- | Same as 'txt', except strip leading whitespace equal to the current
+-- indentation level. Useful for multiline comments, where the AST contains
+-- the full leading whitespace.
+txtStripIndent :: Text -> R ()
+txtStripIndent s = do
+  indent <- R (asks rcIndent)
+  let (leadingSpaces, s') = T.span isSpace s
+  txt $ T.drop indent leadingSpaces <> s'
+
 -- | Similar to 'txt' but the text inserted this way is assumed to break the
 -- “link” between the preceding atom and its pending comments.
 interferingTxt ::
@@ -337,6 +360,23 @@ space = R . modify $ \sc ->
 declNewline :: R ()
 declNewline = newlineRawN =<< getPrinterOpt poNewlinesBetweenDecls
 
+-- | Add a newline in a multiline comment.
+--
+-- Can't use 'newline' because it avoids adding multiple consecutive newlines.
+-- Single-line comments don't have this issue because each line has at least "--".
+multilineCommentNewline :: R ()
+multilineCommentNewline =
+  -- inlining part of rawNewline here because rawNewline still swallows newlines
+  -- when called consecutively
+  R . modify $ \sc ->
+    sc
+      { scBuilder = scBuilder sc <> "\n",
+        scColumn = 0,
+        scIndent = 0,
+        scThisLineSpans = [],
+        scRequestedDelimiter = AfterNewline
+      }
+
 -- | Output a newline. First time 'newline' is used after some non-'newline'
 -- output it gets inserted immediately. Second use of 'newline' does not
 -- output anything but makes sure that the next non-white space output will
@@ -408,8 +448,8 @@ askModuleFixityMap = R (asks rcModuleFixityMap)
 
 -- | Retrieve whether we should print out certain debug information while
 -- printing.
-askDebug :: R Bool
-askDebug = R (asks rcDebug)
+askDebug :: R (Choice "debug")
+askDebug = R (asks (Choice.fromBool . rcDebug))
 
 -- | Like 'inci', but indents by exactly the given number of steps.
 inciBy :: Int -> R () -> R ()
@@ -417,7 +457,37 @@ inciBy step (R m) = R (local modRC m)
   where
     modRC rc =
       rc
-        { rcIndent = roundDownToNearest step (rcIndent rc) + step
+        { rcIndent =
+            -- Currently, in parenthesized expressions, we keep expressions aligned on
+            -- the left, and increase indentation from that point. But this means that
+            -- all subsequent indentation levels will be misaligned with the indentation,
+            -- assuming the parentheses starts on the indentation column:
+            --
+            --   ( Maybe
+            --         Int ->
+            --     String
+            --   )
+            --     ^^^^ 'Int' is 4 spaces from indentation start
+            --   ^^^^^^ but 6 spaces from start of column
+            --
+            -- To fix this, we mitigated the symptom by rounding down the indentation to
+            -- the indentation first before incrementing.
+            -- https://github.com/fourmolu/fourmolu/pull/37#discussion_r497904191
+            --
+            -- But this causes a regression with 2 space indentation, where a wrapped line
+            -- isn't indented enough if the parentheses does NOT start on the indentation
+            -- column.
+            -- https://github.com/fourmolu/fourmolu/issues/428
+            --
+            -- Ideally, we'd rather indicate that the "current" indentation is shifted,
+            -- while the "next" indentation should increase from the parentheses. However,
+            -- that would probably be a lot of reworking the rcIndent machinery. Instead,
+            -- we'll just restrict this patch to only apply on indentation > 2.
+            let modifyIndent =
+                  if poIndentation (rcPrinterOpts rc) > 2
+                    then roundDownToNearest step
+                    else id
+             in modifyIndent (rcIndent rc) + step
         }
     roundDownToNearest r n = (n `div` r) * r
 
@@ -488,6 +558,9 @@ getLayout = R (asks rcLayout)
 getPrinterOpt :: (forall f. PrinterOpts f -> f a) -> R a
 getPrinterOpt f = R $ asks $ runIdentity . f . rcPrinterOpts
 
+getLocalModules :: R (Set ModuleName)
+getLocalModules = R $ asks rcLocalModules
+
 ----------------------------------------------------------------------------
 -- Special helpers for comment placement
 
@@ -538,6 +611,16 @@ popComment f = R $ do
       return $ Just x
     _ -> return Nothing
 
+-- | Get the comments contained in the enclosing span.
+getEnclosingComments :: R [LComment]
+getEnclosingComments = do
+  isEnclosed <-
+    getEnclosingSpan <&> \case
+      Just enclSpan -> containsSpan enclSpan
+      Nothing -> const False
+  CommentStream cstream <- R $ gets scCommentStream
+  pure $ takeWhile (isEnclosed . getLoc) cstream
+
 -- | Get the immediately enclosing 'RealSrcSpan'.
 getEnclosingSpan :: R (Maybe RealSrcSpan)
 getEnclosingSpan = getEnclosingSpanWhere (const True)
