@@ -72,7 +72,8 @@ data GroupStyle
 p_valDecl :: HsBind GhcPs -> R ()
 p_valDecl = \case
   FunBind _ funId funMatches -> p_funBind funId funMatches
-  PatBind _ pat grhss -> p_match PatternBind False NoSrcStrict [pat] grhss
+  PatBind _ pat multAnn grhss ->
+    p_match PatternBind False multAnn NoSrcStrict [pat] grhss
   VarBind {} -> notImplemented "VarBinds" -- introduced by the type checker
   PatSynBind _ psb -> p_patSynBind psb
 
@@ -89,7 +90,7 @@ p_matchGroup ::
 p_matchGroup = p_matchGroup' exprPlacement p_hsExpr
 
 p_matchGroup' ::
-  ( Anno (GRHS GhcPs (LocatedA body)) ~ SrcAnn NoEpAnns,
+  ( Anno (GRHS GhcPs (LocatedA body)) ~ EpAnnCO,
     Anno (Match GhcPs (LocatedA body)) ~ SrcSpanAnnA
   ) =>
   -- | How to get body placement
@@ -119,6 +120,7 @@ p_matchGroup' placer render style mg@MG {..} = do
         render
         (adjustMatchGroupStyle m style)
         (isInfixMatch m)
+        (HsNoMultAnn NoExtField)
         (matchStrictness m)
         m_pats
         m_grhss
@@ -148,6 +150,8 @@ p_match ::
   MatchGroupStyle ->
   -- | Is this an infix match?
   Bool ->
+  -- | Multiplicity annotation
+  HsMultAnn GhcPs ->
   -- | Strictness prefix (FunBind)
   SrcStrictness ->
   -- | Argument patterns
@@ -158,7 +162,7 @@ p_match ::
 p_match = p_match' exprPlacement p_hsExpr
 
 p_match' ::
-  (Anno (GRHS GhcPs (LocatedA body)) ~ SrcAnn NoEpAnns) =>
+  (Anno (GRHS GhcPs (LocatedA body)) ~ EpAnnCO) =>
   -- | How to get body placement
   (body -> Placement) ->
   -- | How to print body
@@ -167,6 +171,8 @@ p_match' ::
   MatchGroupStyle ->
   -- | Is this an infix match?
   Bool ->
+  -- | Multiplicity annotation
+  HsMultAnn GhcPs ->
   -- | Strictness prefix (FunBind)
   SrcStrictness ->
   -- | Argument patterns
@@ -174,7 +180,7 @@ p_match' ::
   -- | Equations
   GRHSs GhcPs (LocatedA body) ->
   R ()
-p_match' placer render style isInfix strictness m_pats GRHSs {..} = do
+p_match' placer render style isInfix multAnn strictness m_pats GRHSs {..} = do
   -- Normally, since patterns may be placed in a multi-line layout, it is
   -- necessary to bump indentation for the pattern group so it's more
   -- indented than function name. This in turn means that indentation for
@@ -182,6 +188,13 @@ p_match' placer render style isInfix strictness m_pats GRHSs {..} = do
   -- would start with two indentation steps applied, which is ugly, so we
   -- need to be a bit more clever here and bump indentation level only when
   -- pattern group is multiline.
+  case multAnn of
+    HsNoMultAnn NoExtField -> pure ()
+    HsPct1Ann _ -> txt "%1" *> space
+    HsMultAnn _ ty -> do
+      txt "%"
+      located ty p_hsType
+      space
   case strictness of
     NoSrcStrict -> return ()
     SrcStrict -> txt "!"
@@ -213,6 +226,7 @@ p_match' placer render style isInfix strictness m_pats GRHSs {..} = do
                   LazyPat _ _ -> True
                   BangPat _ _ -> True
                   SplicePat _ _ -> True
+                  InvisPat _ _ -> True
                   _ -> False
             txt "\\"
             when needsSpace space
@@ -363,19 +377,17 @@ p_hsCmd' isApp s = \case
     located cmd (p_hsCmd' Applicand s)
     breakpoint
     inci $ located expr p_hsExpr
-  HsCmdLam _ mgroup -> p_matchGroup' cmdPlacement p_hsCmd Lambda mgroup
-  HsCmdPar _ _ c _ -> parens N $ sitcc $ located c p_hsCmd
+  HsCmdLam _ variant mgroup -> p_lam isApp variant cmdPlacement p_hsCmd mgroup
+  HsCmdPar _ c -> parens N $ sitcc $ located c p_hsCmd
   HsCmdCase _ e mgroup ->
     p_case isApp cmdPlacement p_hsCmd e mgroup
-  HsCmdLamCase _ variant mgroup ->
-    p_lamcase isApp variant cmdPlacement p_hsCmd mgroup
   HsCmdIf anns _ if' then' else' ->
     p_if cmdPlacement p_hsCmd anns if' then' else'
-  HsCmdLet _ letToken localBinds _ c ->
+  HsCmdLet (letToken, _) localBinds c ->
     p_let (s == S) p_hsCmd letToken localBinds c
   HsCmdDo _ es -> do
     txt "do"
-    p_stmts isApp cmdPlacement (p_hsCmd' NotApplicand S) es
+    p_stmts S isApp cmdPlacement (p_hsCmd' NotApplicand) es
 
 -- | Print a top-level command.
 p_hsCmdTop :: BracketStyle -> HsCmdTop GhcPs -> R ()
@@ -411,21 +423,22 @@ withSpacing f l = located l $ \x -> do
         _ -> setSpanMark (StatementSpan currentSpn)
 
 p_stmt :: Stmt GhcPs (LHsExpr GhcPs) -> R ()
-p_stmt = p_stmt' exprPlacement p_hsExpr
+p_stmt = p_stmt' N exprPlacement (p_hsExpr' NotApplicand)
 
 p_stmt' ::
   ( Anno (Stmt GhcPs (LocatedA body)) ~ SrcSpanAnnA,
     Anno [LocatedA (Stmt GhcPs (LocatedA body))] ~ SrcSpanAnnL
   ) =>
+  BracketStyle ->
   -- | Placer
   (body -> Placement) ->
   -- | Render
-  (body -> R ()) ->
+  (BracketStyle -> body -> R ()) ->
   -- | Statement to render
   Stmt GhcPs (LocatedA body) ->
   R ()
-p_stmt' placer render = \case
-  LastStmt _ body _ _ -> located body render
+p_stmt' s placer render = \case
+  LastStmt _ body _ _ -> located body (render s)
   BindStmt _ p f@(getLocA -> l) -> do
     located p p_pat
     space
@@ -435,25 +448,24 @@ p_stmt' placer render = \case
           | isOneLineSpan (mkSrcSpan (srcSpanEnd loc) (srcSpanStart l)) = placer (unLoc f)
           | otherwise = Normal
     switchLayout [loc, l] $
-      placeHanging placement (located f render)
+      placeHanging placement (located f (render N))
   ApplicativeStmt {} -> notImplemented "ApplicativeStmt" -- generated by renamer
-  BodyStmt _ body _ _ -> located body render
+  BodyStmt _ body _ _ -> located body (render s)
   LetStmt epAnnLet binds -> do
     let letLoc =
           fmap (\(AddEpAnn _ loc) -> loc)
             . find (\(AddEpAnn ann _) -> ann == AnnLet)
-            . epAnnAnns
             $ epAnnLet
     p_let' True letLoc binds Nothing
   ParStmt {} ->
-    -- 'ParStmt' should always be eliminated in 'gatherStmt' already, such
+    -- 'ParStmt' should always be eliminated in 'gatherStmts' already, such
     -- that it never occurs in 'p_stmt''. Consequently, handling it here
     -- would be redundant.
     notImplemented "ParStmt"
   TransStmt {..} ->
     -- 'TransStmt' only needs to account for render printing itself, since
     -- pretty printing of relevant statements (e.g., in 'trS_stmts') is
-    -- handled through 'gatherStmt'.
+    -- handled through 'gatherStmts'.
     case (trS_form, trS_by) of
       (ThenForm, Nothing) -> do
         txt "then"
@@ -482,25 +494,26 @@ p_stmt' placer render = \case
   RecStmt {..} -> do
     txt "rec"
     space
-    sitcc . located recS_stmts $ sepSemi (withSpacing (p_stmt' placer render))
+    sitcc . located recS_stmts $ sepSemi (withSpacing (p_stmt' s placer render))
 
 p_stmts ::
   ( Anno (Stmt GhcPs (LocatedA body)) ~ SrcSpanAnnA,
     Anno [LocatedA (Stmt GhcPs (LocatedA body))] ~ SrcSpanAnnL
   ) =>
+  BracketStyle ->
   IsApplicand ->
   -- | Placer
   (body -> Placement) ->
   -- | Render
-  (body -> R ()) ->
+  (BracketStyle -> body -> R ()) ->
   -- | Statements to render
   LocatedL [LocatedA (Stmt GhcPs (LocatedA body))] ->
   R ()
-p_stmts isApp placer render es = do
+p_stmts s isApp placer render es = do
   breakpoint
   ub <- layoutToBraces <$> getLayout
   let p_stmtExt (relPos, stmt) =
-        ub' $ withSpacing (p_stmt' placer render) stmt
+        ub' $ withSpacing (p_stmt' s placer render) stmt
         where
           -- We need to set brace usage information for all but the last
           -- statement (e.g.in the case of nested do blocks).
@@ -512,17 +525,6 @@ p_stmts isApp placer render es = do
   inciApplicand isApp . located es $
     sepSemi p_stmtExt . attachRelativePos
 
-gatherStmt :: ExprLStmt GhcPs -> [[ExprLStmt GhcPs]]
-gatherStmt (L _ (ParStmt _ block _ _)) =
-  foldr ((<>) . gatherStmtBlock) [] block
-gatherStmt (L s stmt@TransStmt {..}) =
-  foldr liftAppend [] ((gatherStmt <$> trS_stmts) <> pure [[L s stmt]])
-gatherStmt stmt = [[stmt]]
-
-gatherStmtBlock :: ParStmtBlock GhcPs GhcPs -> [[ExprLStmt GhcPs]]
-gatherStmtBlock (ParStmtBlock _ stmts _ _) =
-  foldr (liftAppend . gatherStmt) [] stmts
-
 p_hsLocalBinds :: HsLocalBinds GhcPs -> R ()
 p_hsLocalBinds = \case
   HsValBinds epAnn (ValBinds _ bag lsigs) -> pseudoLocated epAnn $ do
@@ -559,11 +561,10 @@ p_hsLocalBinds = \case
     -- of p_hsLocalBinds). Hence, we introduce a manual Located as we
     -- depend on the layout being correctly set.
     pseudoLocated = \case
-      EpAnn {anns = AnnList {al_anchor = Just Anchor {anchor}}}
-        | let sp = RealSrcSpan anchor Strict.Nothing,
-          -- excluding cases where there are no bindings
-          not $ isZeroWidthSpan sp ->
-            located (L sp ()) . const
+      EpAnn {anns = AnnList {al_anchor}}
+        | -- excluding cases where there are no bindings
+          not $ isZeroWidthSpan (locA al_anchor) ->
+            located (L al_anchor ()) . const
       _ -> id
 
 p_ldotFieldOcc :: XRec GhcPs (DotFieldOcc GhcPs) -> R ()
@@ -577,7 +578,7 @@ p_fieldOcc :: FieldOcc GhcPs -> R ()
 p_fieldOcc FieldOcc {..} = p_rdrName foLabel
 
 p_hsFieldBind ::
-  (lhs ~ GenLocated l a, HasSrcSpan l) =>
+  (lhs ~ GenLocated l a, HasLoc l) =>
   (lhs -> R ()) ->
   HsFieldBind lhs (LHsExpr GhcPs) ->
   R ()
@@ -587,7 +588,7 @@ p_hsFieldBind p_lhs HsFieldBind {..} = do
     space
     equals
     let placement =
-          if onTheSameLine (getLoc' hfbLHS) (getLocA hfbRHS)
+          if onTheSameLine (getLocA hfbLHS) (getLocA hfbRHS)
             then exprPlacement (unLoc hfbRHS)
             else Normal
     placeHanging placement (located hfbRHS p_hsExpr)
@@ -626,10 +627,8 @@ p_hsExpr' isApp s = \case
       HsString (SourceText stxt) _ -> p_stringLit stxt
       HsStringPrim (SourceText stxt) _ -> p_stringLit stxt
       r -> atom r
-  HsLam _ mgroup ->
-    p_matchGroup Lambda mgroup
-  HsLamCase _ variant mgroup ->
-    p_lamcase isApp variant exprPlacement p_hsExpr mgroup
+  HsLam _ variant mgroup ->
+    p_lam isApp variant exprPlacement p_hsExpr mgroup
   HsApp _ f x -> do
     let -- In order to format function applications with multiple parameters
         -- nicer, traverse the AST to gather the function and all the
@@ -687,7 +686,7 @@ p_hsExpr' isApp s = \case
           sep breakpoint (located' p_hsExpr) initp
         placeHanging placement $
           located lastp p_hsExpr
-  HsAppType _ e _ a -> do
+  HsAppType _ e a -> do
     located e p_hsExpr
     breakpoint
     inci $ do
@@ -716,8 +715,11 @@ p_hsExpr' isApp s = \case
     -- negated literals, as `- 1` and `-1` have differing AST.
     when (negativeLiterals && isLiteral) space
     located e p_hsExpr
-  HsPar _ _ e _ ->
-    parens s $ sitcc (located e (dontUseBraces . p_hsExpr))
+  HsPar _ e -> do
+    csSpans <-
+      fmap (flip RealSrcSpan Strict.Nothing . getLoc) <$> getEnclosingComments
+    switchLayout (locA e : csSpans) $
+      parens s (sitcc $ located e (dontUseBraces . p_hsExpr))
   SectionL _ x op -> do
     located x p_hsExpr
     breakpoint
@@ -759,36 +761,18 @@ p_hsExpr' isApp s = \case
     txt "if"
     breakpoint
     inciApplicand isApp $ sep newline (located' (p_grhs RightArrow)) guards
-  HsLet _ letToken localBinds _ e ->
+  HsLet (letToken, _) localBinds e ->
     p_let (s == S) p_hsExpr letToken localBinds e
   HsDo _ doFlavor es -> do
     let doBody moduleName header = do
           forM_ moduleName $ \m -> atom m *> txt "."
           txt header
-          p_stmts isApp exprPlacement (p_hsExpr' NotApplicand S) es
-        compBody = brackets s . located es $ \xs -> do
-          let p_parBody =
-                sep
-                  (breakpoint >> txt "|" >> space)
-                  p_seqBody
-              p_seqBody =
-                sitccIfTrailing
-                  . sep
-                    commaDel
-                    (located' (sitcc . p_stmt))
-              stmts = init xs
-              yield = last xs
-              lists = foldr (liftAppend . gatherStmt) [] stmts
-          located yield p_stmt
-          breakpoint
-          txt "|"
-          space
-          p_parBody lists
+          p_stmts S isApp exprPlacement (p_hsExpr' NotApplicand) es
     case doFlavor of
       DoExpr moduleName -> doBody moduleName "do"
       MDoExpr moduleName -> doBody moduleName "mdo"
-      ListComp -> compBody
-      MonadComp -> compBody
+      ListComp -> p_listComp s es
+      MonadComp -> p_listComp s es
       GhciStmtCtxt -> notImplemented "GhciStmtCtxt"
   ExplicitList _ xs ->
     brackets s $
@@ -857,7 +841,7 @@ p_hsExpr' isApp s = \case
     located expr p_hsExpr
     breakpoint'
     txt "||]"
-  HsUntypedBracket epAnn x -> p_hsQuote epAnn x
+  HsUntypedBracket anns x -> p_hsQuote anns x
   HsTypedSplice _ expr -> p_hsSpliceTH True expr DollarSplice
   HsUntypedSplice _ untySplice -> p_hsUntypedSplice DollarSplice untySplice
   HsProc _ p e -> do
@@ -881,6 +865,122 @@ p_hsExpr' isApp s = \case
       breakpoint
       let inciIfS = case s of N -> id; S -> inci
       inciIfS $ located x p_hsExpr
+  HsEmbTy _ HsWC {hswc_body} -> do
+    txt "type"
+    space
+    located hswc_body p_hsType
+
+-- | Print a list comprehension.
+--
+-- BracketStyle should be N except in a do-block, which must be S or else it's a parse error.
+p_listComp :: BracketStyle -> GenLocated SrcSpanAnnL [ExprLStmt GhcPs] -> R ()
+p_listComp s es = sitcc (vlayout singleLine multiLine)
+  where
+    singleLine = do
+      txt "["
+      body
+      txt "]"
+    multiLine = do
+      txt "[" >> space
+      (if s == S then sitcc else id) $ do
+        body
+        newline
+        txt "]"
+
+    body = located es p_body
+    p_body xs = do
+      let (stmts, yield) =
+            -- TODO: use unsnoc when require GHC 9.8+
+            case xs of
+              [] -> error $ "list comprehension unexpectedly had no expressions"
+              _ -> (init xs, last xs)
+      sitcc $ located yield p_stmt
+      breakpoint
+      txt "|"
+      space
+      p_bodyParallels (gatherStmts stmts)
+
+    -- print the list of list comprehension sections, e.g.
+    -- [ "| x <- xs, y <- ys, let z = x <> y", "| a <- f z" ]
+    p_bodyParallels = sep (breakpoint >> txt "|" >> space) (sitccIfTrailing . p_bodyParallelStmts)
+
+    -- print a list comprehension section within a pipe, e.g.
+    -- [ "x <- xs", "y <- ys", "let z = x <> y" ]
+    p_bodyParallelStmts = sep commaDel (located' (sitcc . p_stmt))
+
+-- | Gather the set of statements in a list comprehension.
+--
+-- For example, this code:
+--
+-- @
+-- [ a + b + c + d
+-- | a <- as, let b = a + a
+-- | c <- cs
+-- | d <- ds, then sort by f
+-- ]
+-- @
+--
+-- is parsed as roughly:
+--
+-- @
+-- [ ParStmt
+--     [ ParStmtBlock
+--         [ BindStmt [| a <- as |]
+--         , LetStmt  [| let b = a + a |]
+--         ]
+--     , ParStmtBlock
+--         [ BindStmt [| c <- cs |]
+--         ]
+--     , ParStmtBlock
+--         [ TransStmt
+--             [ BindStmt [| d <- ds |]
+--             ]
+--             [| then sort by f |]
+--         ]
+--     ]
+-- , LastStmt [| a + b + c + d |]
+-- ]
+-- @
+--
+-- The final expression is parsed out in p_body, and the rest is passed
+-- to this function. This function takes the above tree as input and
+-- normalizes it into:
+--
+-- @
+-- [ [ BindStmt [| a <- as |]
+--   , LetStmt  [| let b = a + a |]
+--   ]
+-- , [ BindStmt [| c <- cs |]
+--   ]
+-- , [ BindStmt [| d <- ds |]
+--   , TransStmt [] [| then sortWith by f |]
+--   ]
+-- ]
+-- @
+--
+-- Notes:
+--   * The number of elements in the outer list is the number of pipes in
+--     the comprehension; i.e. 1 unless -XParallelListComp is enabled
+gatherStmts :: [ExprLStmt GhcPs] -> [[ExprLStmt GhcPs]]
+gatherStmts = \case
+  -- When -XParallelListComp is enabled + list comprehension has
+  -- multiple pipes, input will have exactly 1 element, and it
+  -- will be ParStmt.
+  [L _ (ParStmt _ blocks _ _)] ->
+    [ concatMap collectNonParStmts stmts
+    | ParStmtBlock _ stmts _ _ <- blocks
+    ]
+  -- Otherwise, list will not contain any ParStmt
+  stmts ->
+    [ concatMap collectNonParStmts stmts
+    ]
+  where
+    collectNonParStmts = \case
+      L _ ParStmt {} -> unexpected "ParStmt"
+      stmt@(L _ TransStmt {trS_stmts}) -> concatMap collectNonParStmts trS_stmts ++ [stmt]
+      stmt -> [stmt]
+
+    unexpected label = error $ "Unexpected " <> label <> "! Please file a bug."
 
 p_patSynBind :: PatSynBind GhcPs GhcPs -> R ()
 p_patSynBind PSB {..} = do
@@ -942,7 +1042,7 @@ p_patSynBind PSB {..} = do
       inci (rhs conSpans)
 
 p_case ::
-  ( Anno (GRHS GhcPs (LocatedA body)) ~ SrcAnn NoEpAnns,
+  ( Anno (GRHS GhcPs (LocatedA body)) ~ EpAnnCO,
     Anno (Match GhcPs (LocatedA body)) ~ SrcSpanAnnA
   ) =>
   IsApplicand ->
@@ -964,13 +1064,13 @@ p_case isApp placer render e mgroup = do
   breakpoint
   inciApplicand isApp (p_matchGroup' placer render Case mgroup)
 
-p_lamcase ::
-  ( Anno (GRHS GhcPs (LocatedA body)) ~ SrcAnn NoEpAnns,
+p_lam ::
+  ( Anno (GRHS GhcPs (LocatedA body)) ~ EpAnnCO,
     Anno (Match GhcPs (LocatedA body)) ~ SrcSpanAnnA
   ) =>
   IsApplicand ->
-  -- | Variant (@\\case@ or @\\cases@)
-  LamCaseVariant ->
+  -- | Variant (@\\@ or @\\case@ or @\\cases@)
+  HsLamVariant ->
   -- | Placer
   (body -> Placement) ->
   -- | Render
@@ -978,12 +1078,19 @@ p_lamcase ::
   -- | Expression
   MatchGroup GhcPs (LocatedA body) ->
   R ()
-p_lamcase isApp variant placer render mgroup = do
-  txt $ case variant of
-    LamCase -> "\\case"
-    LamCases -> "\\cases"
-  breakpoint
-  inciApplicand isApp (p_matchGroup' placer render LambdaCase mgroup)
+p_lam isApp variant placer render mgroup = do
+  let mCaseTxt = case variant of
+        LamSingle -> Nothing
+        LamCase -> Just "\\case"
+        LamCases -> Just "\\cases"
+      mgs = if isJust mCaseTxt then LambdaCase else Lambda
+      pMatchGroup = p_matchGroup' placer render mgs mgroup
+  case mCaseTxt of
+    Nothing -> pMatchGroup
+    Just caseTxt -> do
+      txt caseTxt
+      breakpoint
+      inciApplicand isApp pMatchGroup
 
 p_if ::
   -- | Placer
@@ -991,7 +1098,7 @@ p_if ::
   -- | Render
   (body -> R ()) ->
   -- | Annotations
-  EpAnn AnnsIf ->
+  AnnsIf ->
   -- | If
   LHsExpr GhcPs ->
   -- | Then
@@ -999,11 +1106,30 @@ p_if ::
   -- | Else
   LocatedA body ->
   R ()
-p_if placer render epAnn if' then' else' = do
+p_if placer render anns if' then' else' = do
   txt "if"
   space
   located if' p_hsExpr
   breakpoint
+  commentSpans <- fmap getLoc <$> getEnclosingComments
+  let (thenSpan, elseSpan) = (locA aiThen, locA aiElse)
+        where
+          AnnsIf {aiThen, aiElse} = anns
+
+      locatedToken tokenSpan token =
+        located (L tokenSpan ()) $ \_ -> txt token
+
+      betweenSpans spanA spanB s = spanA < s && s < spanB
+
+      placeHangingLocated tokenSpan bodyLoc@(L _ body) = do
+        let bodySpan = getLocA bodyLoc
+            hasComments = fromMaybe False $ do
+              tokenRealSpan <- srcSpanToRealSrcSpan tokenSpan
+              bodyRealSpan <- srcSpanToRealSrcSpan bodySpan
+              pure $ any (betweenSpans tokenRealSpan bodyRealSpan) commentSpans
+            placement = if hasComments then Normal else placer body
+        switchLayout [tokenSpan, bodySpan] $
+          placeHanging placement (located bodyLoc render)
   inci $ do
     locatedToken thenSpan "then"
     space
@@ -1012,34 +1138,6 @@ p_if placer render epAnn if' then' else' = do
     locatedToken elseSpan "else"
     space
     placeHangingLocated elseSpan else'
-  where
-    (thenSpan, elseSpan, commentSpans) =
-      case epAnn of
-        EpAnn {anns = AnnsIf {aiThen, aiElse}, comments} ->
-          ( loc' $ epaLocationRealSrcSpan aiThen,
-            loc' $ epaLocationRealSrcSpan aiElse,
-            map (anchor . getLoc) $
-              case comments of
-                EpaComments cs -> cs
-                EpaCommentsBalanced pre post -> pre <> post
-          )
-        EpAnnNotUsed ->
-          (noSrcSpan, noSrcSpan, [])
-
-    locatedToken tokenSpan token =
-      located (L tokenSpan ()) $ \_ -> txt token
-
-    betweenSpans spanA spanB s = spanA < s && s < spanB
-
-    placeHangingLocated tokenSpan bodyLoc@(L _ body) = do
-      let bodySpan = getLoc' bodyLoc
-          hasComments = fromMaybe False $ do
-            tokenRealSpan <- srcSpanToRealSrcSpan tokenSpan
-            bodyRealSpan <- srcSpanToRealSrcSpan bodySpan
-            pure $ any (betweenSpans tokenRealSpan bodyRealSpan) commentSpans
-          placement = if hasComments then Normal else placer body
-      switchLayout [tokenSpan, bodySpan] $
-        placeHanging placement (located bodyLoc render)
 
 p_let ::
   -- | True if in do-block
@@ -1047,16 +1145,16 @@ p_let ::
   -- | Render
   (body -> R ()) ->
   -- | Annotation for the `let` block
-  LHsToken "let" GhcPs ->
+  EpToken "let" ->
   HsLocalBinds GhcPs ->
   LocatedA body ->
   R ()
 p_let inDo render letToken localBinds e = p_let' inDo letLoc localBinds $ Just (located e render)
   where
     letLoc =
-      case getLoc letToken of
-        TokenLoc loc -> Just loc
-        NoTokenLoc -> Nothing
+      case letToken of
+        EpTok loc -> Just loc
+        NoEpTok -> Nothing
 
 p_let' ::
   -- | True if in do-block
@@ -1132,9 +1230,10 @@ p_let' inDo letLoc localBinds mBody = do
       HsValBinds epanns _ -> Just epanns
       HsIPBinds epanns _ -> Just epanns
       EmptyLocalBinds _ -> Nothing
-    epAnnsStartLine = \case
-      EpAnn {entry} -> Just (srcSpanStartLine . anchor $ entry)
-      EpAnnNotUsed -> Nothing
+    epAnnsStartLine epAnn =
+      case entry epAnn of
+        EpaSpan (RealSrcSpan r _) -> Just $ srcSpanStartLine r
+        _ -> Nothing
 
 p_pat :: Pat GhcPs -> R ()
 p_pat = \case
@@ -1143,11 +1242,11 @@ p_pat = \case
   LazyPat _ pat -> do
     txt "~"
     located pat p_pat
-  AsPat _ name _ pat -> do
+  AsPat _ name pat -> do
     p_rdrName name
     txt "@"
     located pat p_pat
-  ParPat _ _ pat _ ->
+  ParPat _ pat ->
     located pat (parens S . sitcc . p_pat)
   BangPat _ pat -> do
     txt "!"
@@ -1212,12 +1311,17 @@ p_pat = \case
   SigPat _ pat HsPS {..} -> do
     located pat p_pat
     p_typeAscription (lhsTypeToSigType hsps_body)
+  EmbTyPat _ (HsTP _ ty) -> do
+    txt "type"
+    space
+    located ty p_hsType
+  InvisPat _ tyPat -> p_tyPat tyPat
 
-p_hsPatSigType :: HsPatSigType GhcPs -> R ()
-p_hsPatSigType (HsPS _ ty) = txt "@" *> located ty p_hsType
+p_tyPat :: HsTyPat GhcPs -> R ()
+p_tyPat (HsTP _ ty) = txt "@" *> located ty p_hsType
 
 p_hsConPatTyArg :: HsConPatTyArg GhcPs -> R ()
-p_hsConPatTyArg (HsConPatTyArg _ patSigTy) = p_hsPatSigType patSigTy
+p_hsConPatTyArg (HsConPatTyArg _ patSigTy) = p_tyPat patSigTy
 
 p_pat_hsFieldBind :: HsRecField GhcPs (LPat GhcPs) -> R ()
 p_pat_hsFieldBind HsFieldBind {..} = do
@@ -1272,11 +1376,11 @@ p_hsSpliceTH isTyped expr = \case
   where
     decoSymbol = if isTyped then "$$" else "$"
 
-p_hsQuote :: EpAnn [AddEpAnn] -> HsQuote GhcPs -> R ()
-p_hsQuote epAnn = \case
+p_hsQuote :: [AddEpAnn] -> HsQuote GhcPs -> R ()
+p_hsQuote anns = \case
   ExpBr _ expr -> do
     let name
-          | any isJust (matchAddEpAnn AnnOpenEQ <$> epAnnAnns epAnn) = ""
+          | any (isJust . matchAddEpAnn AnnOpenEQ) anns = ""
           | otherwise = "e"
     quote name (located expr p_hsExpr)
   PatBr _ pat -> located pat (quote "p" . p_pat)
@@ -1364,14 +1468,6 @@ layoutToBraces = \case
   SingleLine -> useBraces
   MultiLine -> id
 
--- | Append each element in both lists with semigroups. If one list is shorter
--- than the other, return the rest of the longer list unchanged.
-liftAppend :: (Semigroup a) => [a] -> [a] -> [a]
-liftAppend [] [] = []
-liftAppend [] (y : ys) = y : ys
-liftAppend (x : xs) [] = x : xs
-liftAppend (x : xs) (y : ys) = x <> y : liftAppend xs ys
-
 getGRHSSpan :: GRHS GhcPs (LocatedA body) -> SrcSpan
 getGRHSSpan (GRHS _ guards body) =
   combineSrcSpans' $ getLocA body :| map getLocA guards
@@ -1387,10 +1483,9 @@ blockPlacement _ _ = Normal
 -- | Determine placement of a given command.
 cmdPlacement :: HsCmd GhcPs -> Placement
 cmdPlacement = \case
-  HsCmdLam _ _ -> Hanging
-  HsCmdCase _ _ _ -> Hanging
-  HsCmdLamCase _ _ _ -> Hanging
-  HsCmdDo _ _ -> Hanging
+  HsCmdLam {} -> Hanging
+  HsCmdCase {} -> Hanging
+  HsCmdDo {} -> Hanging
   _ -> Normal
 
 -- | Determine placement of a top level command.
@@ -1401,12 +1496,14 @@ cmdTopPlacement (HsCmdTop _ (L _ x)) = cmdPlacement x
 exprPlacement :: HsExpr GhcPs -> Placement
 exprPlacement = \case
   -- Only hang lambdas with single line parameter lists
-  HsLam _ mg -> case mg of
-    MG _ (L _ [L _ (Match _ _ (x : xs) _)])
-      | isOneLineSpan (combineSrcSpans' $ fmap getLocA (x :| xs)) ->
-          Hanging
-    _ -> Normal
-  HsLamCase _ _ _ -> Hanging
+  HsLam _ variant mg -> case variant of
+    LamSingle -> case mg of
+      MG _ (L _ [L _ (Match _ _ (x : xs) _)])
+        | isOneLineSpan (combineSrcSpans' $ fmap getLocA (x :| xs)) ->
+            Hanging
+      _ -> Normal
+    LamCase -> Hanging
+    LamCases -> Hanging
   HsCase _ _ _ -> Hanging
   HsDo _ (DoExpr _) _ -> Hanging
   HsDo _ (MDoExpr _) _ -> Hanging
