@@ -1,5 +1,7 @@
 {-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE OverloadedLabels #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE PatternSynonyms #-}
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE TypeFamilies #-}
 
@@ -22,7 +24,10 @@ module Ormolu.Printer.Meat.Type
 where
 
 import Control.Monad
+import Data.Choice (pattern With, pattern Without)
 import Data.Functor ((<&>))
+import Data.List (sortOn)
+import GHC.Data.Strict qualified as Strict
 import GHC.Hs hiding (isPromoted)
 import GHC.Types.SourceText
 import GHC.Types.SrcLoc
@@ -50,7 +55,7 @@ p_hsType' multilineArgs = \case
     getPrinterOpt poFunctionArrows >>= \case
       LeadingArrows | multilineArgs -> interArgBreak >> txt " " >> p_forallBndrsEnd vis
       _ -> p_forallBndrsEnd vis >> interArgBreak
-    p_hsTypeR (unLoc t)
+    located t p_hsType
   HsQualTy _ qs t -> do
     located qs p_hsContext
     getPrinterOpt poFunctionArrows >>= \case
@@ -59,7 +64,7 @@ p_hsType' multilineArgs = \case
       LeadingArgsArrows -> space >> token'darrow >> interArgBreak
     case unLoc t of
       HsQualTy {} -> p_hsTypeR (unLoc t)
-      HsFunTy {} -> p_hsTypeR (unLoc t)
+      HsFunTy {} -> located t p_hsType
       _ -> located t p_hsTypeR
   HsTyVar _ p n -> do
     case p of
@@ -84,7 +89,7 @@ p_hsType' multilineArgs = \case
       breakpoint
       inci $
         sep breakpoint (located' p_hsType) args
-  HsAppKindTy _ ty _ kd -> sitcc $ do
+  HsAppKindTy _ ty kd -> sitcc $ do
     -- The first argument is the location of the "@..." part. Not 100% sure,
     -- but I think we can ignore it as long as we use 'located' on both the
     -- type and the kind.
@@ -98,7 +103,7 @@ p_hsType' multilineArgs = \case
           case arrow of
             HsUnrestrictedArrow _ -> token'rarrow
             HsLinearArrow _ -> token'lolly
-            HsExplicitMult _ mult _ -> do
+            HsExplicitMult _ mult -> do
               txt "%"
               p_hsTypeR (unLoc mult)
               space
@@ -125,8 +130,11 @@ p_hsType' multilineArgs = \case
     let opTree = BinaryOpBranches (tyOpTree x) op (tyOpTree y)
     p_tyOpTree
       (reassociateOpTree debug (Just . unLoc) modFixityMap opTree)
-  HsParTy _ t ->
-    parens N $ sitcc $ located t p_hsType
+  HsParTy _ t -> do
+    csSpans <-
+      fmap (flip RealSrcSpan Strict.Nothing . getLoc) <$> getEnclosingComments
+    switchLayout (locA t : csSpans) $
+      parens N (sitcc $ located t p_hsType)
   HsIParamTy _ n t -> sitcc $ do
     located n atom
     inci $ startTypeAnnotation t p_hsType
@@ -143,12 +151,12 @@ p_hsType' multilineArgs = \case
         LeadingArgsArrows -> False
     if usePipe
       then do
-        p_hsDoc Pipe True str
+        p_hsDoc Pipe (With #endNewline) str
         located t p_hsType
       else do
         located t p_hsType
         newline
-        p_hsDoc Caret False str
+        p_hsDoc Caret (Without #endNewline) str
   HsBangTy _ (HsSrcBang _ u s) t -> do
     case u of
       SrcUnpack -> txt "{-# UNPACK #-}" >> space
@@ -200,14 +208,14 @@ p_hsType' multilineArgs = \case
     p_hsTypeR m = p_hsType' multilineArgs m
 
 startTypeAnnotation ::
-  (HasSrcSpan l) =>
+  (HasLoc l) =>
   GenLocated l a ->
   (a -> R ()) ->
   R ()
 startTypeAnnotation = startTypeAnnotation' breakpoint breakpoint
 
 startTypeAnnotationDecl ::
-  (HasSrcSpan l) =>
+  (HasLoc l) =>
   GenLocated l a ->
   (a -> HsType GhcPs) ->
   (a -> R ()) ->
@@ -222,7 +230,7 @@ startTypeAnnotationDecl lItem getType =
     lItem
 
 startTypeAnnotation' ::
-  (HasSrcSpan l) =>
+  (HasLoc l) =>
   R () ->
   R () ->
   GenLocated l a ->
@@ -261,7 +269,10 @@ p_hsContext :: HsContext GhcPs -> R ()
 p_hsContext = \case
   [] -> txt "()"
   [x] -> located x p_hsType
-  xs -> parens N $ sep commaDel (sitcc . located' p_hsType) xs
+  xs -> do
+    shouldSort <- getPrinterOpt poSortConstraints
+    let sort = if shouldSort then sortOn showOutputable else id
+    parens N $ sep commaDel (sitcc . located' p_hsType) (sort xs)
 
 class IsTyVarBndrFlag flag where
   isInferred :: flag -> Bool
@@ -279,7 +290,7 @@ instance IsTyVarBndrFlag Specificity where
 instance IsTyVarBndrFlag (HsBndrVis GhcPs) where
   isInferred _ = False
   p_tyVarBndrFlag = \case
-    HsBndrRequired -> pure ()
+    HsBndrRequired NoExtField -> pure ()
     HsBndrInvisible _ -> txt "@"
 
 p_hsTyVarBndr :: (IsTyVarBndrFlag flag) => HsTyVarBndr flag GhcPs -> R ()
@@ -297,7 +308,7 @@ data ForAllVisibility = ForAllInvis | ForAllVis
 
 -- | Render several @forall@-ed variables.
 p_forallBndrs ::
-  (HasSrcSpan l) =>
+  (HasLoc l) =>
   ForAllVisibility ->
   (a -> R ()) ->
   [GenLocated l a] ->
@@ -306,10 +317,10 @@ p_forallBndrs vis p tyvars = do
   p_forallBndrsStart p tyvars
   p_forallBndrsEnd vis
 
-p_forallBndrsStart :: (HasSrcSpan l) => (a -> R ()) -> [GenLocated l a] -> R ()
+p_forallBndrsStart :: (HasLoc l) => (a -> R ()) -> [GenLocated l a] -> R ()
 p_forallBndrsStart _ [] = token'forall
 p_forallBndrsStart p tyvars = do
-  switchLayout (getLoc' <$> tyvars) $ do
+  switchLayout (locA <$> tyvars) $ do
     token'forall
     breakpoint
     inci $ do
@@ -327,7 +338,7 @@ p_conDeclField :: ConDeclField GhcPs -> R ()
 p_conDeclField ConDeclField {..} = do
   commaStyle <- getPrinterOpt poCommaStyle
   when (commaStyle == Trailing) $
-    mapM_ (p_hsDoc Pipe True) cd_fld_doc
+    mapM_ (p_hsDoc Pipe (With #endNewline)) cd_fld_doc
   sitcc $
     sep
       commaDel
@@ -350,11 +361,11 @@ p_conDeclField ConDeclField {..} = do
       breakpoint
       sitcc . inci $ p_hsType (unLoc cd_fld_type)
   when (commaStyle == Leading) $
-    mapM_ (inciByFrac (-1) . (newline >>) . p_hsDoc Caret False) cd_fld_doc
+    mapM_ (inciByFrac (-1) . (newline >>) . p_hsDoc Caret (Without #endNewline)) cd_fld_doc
 
 p_lhsTypeArg :: LHsTypeArg GhcPs -> R ()
 p_lhsTypeArg = \case
-  HsValArg ty -> located ty p_hsType
+  HsValArg NoExtField ty -> located ty p_hsType
   -- first argument is the SrcSpan of the @,
   -- but the @ always has to be directly before the type argument
   HsTypeArg _ ty -> txt "@" *> located ty p_hsType
@@ -376,8 +387,8 @@ hsOuterTyVarBndrsToHsType ::
 hsOuterTyVarBndrsToHsType obndrs ty = case obndrs of
   HsOuterImplicit NoExtField -> unLoc ty
   HsOuterExplicit _ bndrs ->
-    HsForAllTy NoExtField (mkHsForAllInvisTele EpAnnNotUsed bndrs) ty
+    HsForAllTy NoExtField (mkHsForAllInvisTele noAnn bndrs) ty
 
 lhsTypeToSigType :: LHsType GhcPs -> LHsSigType GhcPs
 lhsTypeToSigType ty =
-  reLocA . L (getLocA ty) . HsSig NoExtField (HsOuterImplicit NoExtField) $ ty
+  L (getLoc ty) . HsSig NoExtField (HsOuterImplicit NoExtField) $ ty
